package org.ioc.commons.model.globaldata;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.ioc.commons.integration.service.ServiceCallback;
import org.ioc.commons.ui.HasLoader;

/**
 * A tool class for controling the global data from the entire application.
 * 
 * @author 'Jesús Lunar Pérez'
 * 
 * @param <G>
 *            Global data type
 */
public class GlobalDataController<G extends Enum<? extends IsGlobalData>> {
	private enum GlobalDataValueStatus {
		NotRead, Fetching, Cached
	}

	private class StoreEntry {
		private Provider<?> provider;
		private Object value;
		private GlobalDataValueStatus status = GlobalDataValueStatus.NotRead;
		private boolean fireOnUpdate;
		private boolean fireOnReady;
	}

	public interface Provider<V> {

		public void provide(HasLoader globalLoadingIndicator, final ServiceCallback<V> callback);
	}

	public interface Consumer<G extends Enum<? extends IsGlobalData>> {

		void requires(G globalData);

		void ensureRequirements();

	}

	private enum ConsumerStatus {
		NotReady, Ready
	}

	private class ConsumerImpl implements Consumer<G> {

		private Set<G> requirements = new HashSet<G>();
		private IsGlobalDataConsumer<G> forThisObject;
		private ConsumerStatus lastStatus;

		public ConsumerImpl(IsGlobalDataConsumer<G> forThisObject) {
			this.forThisObject = forThisObject;
		}

		@Override
		public void requires(G globalData) {
			// TODO: To throw custom exception
			/**
			 * When a provider are not registered, it will throw an exception
			 */
			assert (globalDataStore.containsKey(globalData)) : "There is no providers registered for " + globalData
					+ ".";

			StoreEntry storeEntry = globalDataStore.get(globalData);

			/**
			 * If there is a registered provider but the value has not been
			 * read, we start the fetching
			 */
			if (storeEntry.status == GlobalDataValueStatus.NotRead) {
				fetchData(globalData);
			}

			this.requirements.add(globalData);
		}

		@Override
		public void ensureRequirements() {
			/**
			 * If ALL the requires are loaded and cached, then it calls
			 * onRequiredGlobalDataReady()
			 */
			/**
			 * In case some one is not loaded yet, it executes
			 * onRequiredGlobalDataNotReadyYet()
			 */
			ConsumerStatus newStatus = ConsumerStatus.Ready;
			for (G requirement : requirements) {
				StoreEntry storeEntry = globalDataStore.get(requirement);
				boolean requirementOk = storeEntry.status == GlobalDataValueStatus.Cached;
				if (!requirementOk) {
					storeEntry.fireOnReady = true;
					newStatus = ConsumerStatus.NotReady;
					break;
				}
			}

			if (lastStatus == null || lastStatus != newStatus) {

				lastStatus = newStatus;

				switch (newStatus) {
				case Ready:
					forThisObject.onRequiredGlobalDataReady();
					break;
				case NotReady:
					forThisObject.onRequiredGlobalDataNotReadyYet();
					break;
				}
			} else {
				/*
				 * The status has not been changed. We are not going to invoke
				 * anything again.
				 */
			}
		}

		boolean isRequiring(G globalData) {
			return this.requirements.contains(globalData);
		}
	}

	public GlobalDataController() {
		this(null);
	}

	public GlobalDataController(HasLoader globalLoadingIndicator) {
		this.globalLoadingIndicator = globalLoadingIndicator;
	}

	private HasLoader globalLoadingIndicator;
	private Map<G, StoreEntry> globalDataStore = new HashMap<G, StoreEntry>();
	private Map<G, Object> globalDataConstants = new HashMap<G, Object>();
	private Map<IsGlobalDataConsumer<G>, ConsumerImpl> registeredConsumers = new HashMap<IsGlobalDataConsumer<G>, ConsumerImpl>();

	public final HasLoader getGlobalLoadingIndicator() {
		return globalLoadingIndicator;
	}

	public final void setGlobalLoadingIndicator(HasLoader globalLoadingIndicator) {
		this.globalLoadingIndicator = globalLoadingIndicator;
	}

	/**
	 * It register a provider for a type of global data.
	 * 
	 * This method is used when the global data may change as opposed to
	 * setConstantGlobalData method.
	 * 
	 * @param globalData
	 * @param provider
	 */
	public <T> void registerGlobalDataProvider(G globalData, Provider<T> provider) {

		// TODO: To throw a custom exception
		assert (!this.globalDataStore.containsKey(globalData)) : "You are trying to register twice a provider for "
				+ globalData;
		assert (!this.globalDataConstants.containsKey(globalData)) : "Cannot register a provider for " + globalData
				+ " when a constant value has been set.";

		StoreEntry se = new StoreEntry();
		se.provider = provider;

		this.globalDataStore.put(globalData, se);
	}

	/**
	 * It set a constant value for a type of global data.
	 * 
	 * This method is used when is known that the global data will never change,
	 * as opposed to registerGlobalDataProvider method.
	 * 
	 * @param globalData
	 *            Global data
	 * @param value
	 *            Value for the global data.
	 */
	public <T> void setGlobalDataConstantValue(G globalData, T value) {

		// TODO: To throw a custom exception
		assert (!this.globalDataStore.containsKey(globalData)) : "Cannot set a constant value for " + globalData
				+ " when a provider has been registered.";

		this.globalDataConstants.put(globalData, value);
	}

	public void refreshGlobalData(final G globalData) {
		// TODO: To throw a custom exception
		assert (this.globalDataStore.containsKey(globalData)) : "There is no providers for " + globalData
				+ ".";

		StoreEntry storeEntry = this.globalDataStore.get(globalData);
		if (storeEntry.status != GlobalDataValueStatus.Fetching) {
			fetchData(globalData);
		}
		storeEntry.fireOnUpdate = true;
	}

	@SuppressWarnings({ "unchecked", "rawtypes" })
	protected void fetchData(final G globalData) {
		StoreEntry storeEntry = this.globalDataStore.get(globalData);
		Provider<?> provider = storeEntry.provider;

		storeEntry.status = GlobalDataValueStatus.Fetching;
		provider.provide(this.globalLoadingIndicator, new ServiceCallback() {

			@Override
			public void onSuccess(Object globalDataValue) {
				StoreEntry storeEntry = globalDataStore.get(globalData);
				storeEntry.value = globalDataValue;
				storeEntry.status = GlobalDataValueStatus.Cached;

				for (ConsumerImpl consumer : registeredConsumers.values()) {
					if (consumer.isRequiring(globalData)) {

						/*
						 * It checks again if all the requeriments are loaded
						 */
						if (storeEntry.fireOnReady) {
							storeEntry.fireOnReady = false;
							consumer.ensureRequirements();
						}

						/*
						 * Moreover, if we were waiting for a refresh, it
						 * executes the method
						 */
						if (storeEntry.fireOnUpdate) {
							storeEntry.fireOnUpdate = false;
							consumer.forThisObject.onRequiredGlobalDataUpdated(globalData);
						}
					}
				}
			}
		});
	}

	public Consumer<G> registerGlobalDataConsumer(IsGlobalDataConsumer<G> forThisObject) {
		ConsumerImpl consumer = this.registeredConsumers.get(forThisObject);
		if (consumer == null) {
			this.registeredConsumers.put(forThisObject, (consumer = new ConsumerImpl(forThisObject)));
		}
		return consumer;
	}

	/**
	 * Get the value from a global data type.
	 * 
	 * In case the global data was registered using a provider, the value will
	 * be got from it the first time and that value <b>will be cached</b>. A
	 * call to refreshGlobalData() will remove the cached value and will fetch
	 * it from the provider again.
	 * 
	 * In case the gloabal data value has been set using
	 * setGlobalDataConstantValue, this value will be returned.
	 * refreshGlobalData() doesn't have any effect in that case.
	 * 
	 * @param globalData
	 * @return Global data value
	 */
	public Object getGlobalDataValue(G globalData) {
		StoreEntry entry = globalDataStore.get(globalData);
		return (entry != null) ? entry.value : (globalDataConstants.get(globalData));
	}
}
